from typing import Any, Tuple, Union, Dict, cast
from abc import ABC

import numpy as np
from scipy import stats
from scipy.optimize import minimize
import pandas as pd

# pymc is optional - just be silent on failed import
try:
    import pymc as pm
except:
    pass


class BasePrior(ABC):
    _rv: stats.distributions.rv_frozen

    def __init__(self, name: str, size: int = 1):
        self.name = name
        self.size = size

    def bounds(self, ci=1.0) -> Tuple[float, float]:
        return self._rv.interval(ci)

    def finite_bounds(self, ci=0.99) -> Tuple[float, float]:
        """Return guaranteed finite bounds, each of the lower and upper bounds
        will be either its full extremum (if finite), or the bound specified by
        the supplied confidence interval

        Args:
            ci: Confidence interval for alternate bounds

        Returns:
            Tuple of lower, upper
        """
        lower, upper = self.bounds(1.0)
        cbounds = self.bounds(ci)
        if np.isinf(lower):
            lower = cbounds[0]
        if np.isinf(upper):
            upper = cbounds[1]
        return lower, upper

    def ppf(self, q):
        """Probability Percentage Function at q (Inverse CDF)
        Defaults to using the underlying scipy distribution function

        Args:
            q: Quantile (float or arraylike) at which to evaluate ppf

        Returns:
            typeof(q): The ppf values
        """

        return self._rv.ppf(q)

    def cdf(self, x):
        """Cumulative Distribution Function at x
        Defaults to using the underlying scipy distribution function

        Args:
            x: Value (float or arraylike) at which to evaluate cdf

        Returns:
            typeof(x): The cdf values
        """
        return self._rv.cdf(x)

    def pdf(self, x):
        """Probability Density Function at x
        Defaults to using the underlying scipy distribution function

        Args:
            x: Value (float or arraylike) at which to evaluate pdf

        Returns:
            typeof(x): The pdf values
        """
        return self._rv.pdf(x)

    def logpdf(self, x):
        """Log Probability Density Function at x
        Defaults to using the underlying scipy distribution function

        Args:
            x: Value (float or arraylike) at which to evaluate logpdf

        Returns:
            typeof(x): The logpdf values
        """
        return self._rv.logpdf(x)

    def get_series(self, func_name, ci=0.99, slen=101):
        x = np.linspace(*self.finite_bounds(ci=ci), slen)
        y = getattr(self._rv, func_name)(x)
        return pd.Series(y, x, name=func_name)

    def _get_pymc_shape(self):
        if self.size == 1:
            return None
        else:
            return self.size

    def __repr__(self):
        return f"{self.__class__.__name__} {self.name}"

    def to_pymc(self):
        raise NotImplementedError()

    @classmethod
    def _get_test(cls):
        raise NotImplementedError()


# Union type for hierarchical/dispersion parameters
DistriParam = Union[float, BasePrior]

# Dict type used by BayesianCompartmentalModel
PriorDict = Dict[str, BasePrior]


class BetaPrior(BasePrior):
    """
    A beta distributed prior.
    """

    def __init__(self, name: str, a: float, b: float, size=1):
        super().__init__(name, size)
        self.a = float(a)
        self.b = float(b)
        self.distri_params = {"a": a, "b": b}
        self._rv = stats.beta(a, b)

    @classmethod
    def from_mean_and_ci(
        cls, name: str, mean: float, ci: Tuple[float, float], ci_width=0.95, size=1
    ):
        assert len(ci) == 2 and ci[1] > ci[0] and 0.0 < ci_width < 1.0
        percentile_low = (1.0 - ci_width) / 2.0
        percentile_up = 1.0 - percentile_low
        assert 0.0 < ci[0] < 1.0 and 0.0 < ci[1] < 1.0 and 0.0 < mean < 1.0

        def distance_to_minimise(a):
            b = a * (1.0 - mean) / mean
            vals = stats.beta.ppf([percentile_low, percentile_up], a, b)
            dist = sum([(ci[i] - vals[i]) ** 2 for i in range(2)])
            return dist

        sol = minimize(distance_to_minimise, [1.0], bounds=[(0.0, None)], tol=1.0e-32)
        best_a = sol.x
        best_b = best_a * (1.0 - mean) / mean

        return cls(name, best_a[0], best_b[0], size)

    def to_pymc(self):
        return pm.Beta(self.name, alpha=self.a, beta=self.b, shape=self._get_pymc_shape())

    @classmethod
    def _get_test(cls):
        return cls("test", 2.0, 5.0)


class UniformPrior(BasePrior):
    """
    A uniformily distributed prior.
    """

    def __init__(self, name: str, domain: Tuple[float, float], size=1):
        super().__init__(name)
        self.start, self.end = domain
        self.distri_params = {"loc": self.start, "scale": self.end - self.start}
        self._rv = stats.uniform(**self.distri_params)
        self.size = size

    def to_pymc(self):
        lower, upper = self.start, self.end
        return pm.Uniform(self.name, lower=lower, upper=upper, shape=self._get_pymc_shape())

    def __repr__(self):
        return f"{super().__repr__()} {{bounds: {self.bounds()}}}"

    @classmethod
    def _get_test(cls):
        return cls("test", (0.0, 1.0))


class TruncNormalPrior(BasePrior):
    """
    A prior with a truncated normal distribution.
    """

    def __init__(
        self, name: str, mean: float, stdev: float, trunc_range: Tuple[float, float], size=1
    ):
        super().__init__(name, size)
        self.mean, self.stdev = mean, stdev
        self.trunc_range = tuple(trunc_range)
        self.distri_params = {
            "loc": mean,
            "scale": stdev,
            "a": (trunc_range[0] - mean) / stdev,
            "b": (trunc_range[1] - mean) / stdev,
        }
        self._rv = stats.truncnorm(**self.distri_params)

    def to_pymc(self):
        lower, upper = self.trunc_range
        return pm.TruncatedNormal(
            self.name,
            mu=self.mean,
            sigma=self.stdev,
            lower=lower,
            upper=upper,
            shape=self._get_pymc_shape(),
        )

    def __repr__(self):
        return f"{super().__repr__()} {{mean: {self.mean}, stdev: {self.stdev}, bounds: {self.bounds()}}}"

    @classmethod
    def _get_test(cls):
        return cls("test", 0.0, 1.0, (0.0, 1.0))


class GammaPrior(BasePrior):
    """A gamma distributed prior"""

    def __init__(self, name: str, shape: float, scale: float, size: int = 1):
        super().__init__(name, size)
        self.shape = float(shape)
        self.scale = float(scale)

        self.distri_params = {"shape": self.shape, "scale": self.scale}

        self._rv = stats.gamma(shape, scale=scale)

    def to_pymc(self):
        alpha = self.shape
        beta = 1.0 / self.scale

        return pm.Gamma(self.name, alpha=alpha, beta=beta, shape=self._get_pymc_shape())

    @classmethod
    def from_mode(
        cls,
        name: str,
        mode: float,
        upper_ci: float,
        size: int = 1,
        tol=1e-6,
        max_eval=8,
        warn=False,
    ):
        def evaluate_gamma(params):
            k, theta = params[0], params[1]
            interval = stats.gamma.interval(0.99, k, scale=theta)
            eval_mode = (k - 1.0) * theta
            return np.abs(eval_mode - mode) + np.abs(interval[-1] - upper_ci)

        x = np.array((1.0, 1.0))
        cur_eval = 0
        loss = np.inf
        while (loss > tol) and (cur_eval < max_eval):
            res = minimize(
                evaluate_gamma, x, bounds=[(1e-8, np.inf), (1e-8, np.inf)], method="Nelder-Mead"
            )
            loss = evaluate_gamma(res.x) / upper_ci
            x = res.x
            cur_eval += 1

        if loss > tol:
            if warn:
                raise RuntimeWarning(
                    f"Loss of {loss} exceeds specified tolerance {tol}, parameters may be impossible"
                )

        return cls(name, x[0], x[1], size)

    @classmethod
    def from_mean(
        cls,
        name: str,
        mean: float,
        upper_ci: float,
        size: int = 1,
        tol=1e-6,
        max_eval=8,
        warn=False,
    ):
        def evaluate_gamma(params):
            k, theta = params[0], params[1]
            interval = stats.gamma.interval(0.99, k, scale=theta)
            eval_mean = stats.gamma.mean(k, scale=theta)
            # Force assumption for static typecheckers
            eval_mean = cast(float, eval_mean)
            return np.abs(eval_mean - mean) + np.abs(interval[-1] - upper_ci)

        x = np.array((1.0, 1.0))
        cur_eval = 0
        loss = np.inf
        while (loss > tol) and (cur_eval < max_eval):
            res = minimize(
                evaluate_gamma, x, bounds=[(1e-8, np.inf), (1e-8, np.inf)], method="Nelder-Mead"
            )
            loss = evaluate_gamma(res.x) / upper_ci
            x = res.x
            cur_eval += 1

        if loss > tol:
            if warn:
                raise RuntimeWarning(
                    f"Loss of {loss} exceeds specified tolerance {tol}, parameters may be impossible"
                )

        return cls(name, x[0], x[1], size)

    @classmethod
    def _get_test(cls):
        return cls("test", 1.0, 0.5)



from typing import Dict

import numpy as np
import pandas as pd

from estival.priors import PriorDict


class SampleTypes:
    INPUT = "input"
    DICT = "dict"
    LIST_OF_DICTS = "list_of_dicts"
    ARRAY = "array"
    PANDAS = "pandas"


def constrain(sample, priors: PriorDict, bounds=0.99):
    # return {k:np.clip(sample[k], *priors[k].bounds(bounds)) for k,v in sample.items()}
    def constrain_op(sample, prior):
        return np.clip(sample, *prior.bounds(bounds))

    return _process_samples_for_priors(sample, priors, constrain_op)


def ppf(sample, priors: PriorDict):
    def ppf_op(sample, prior):
        return prior.ppf(sample)

    return _process_samples_for_priors(sample, priors, ppf_op)


def cdf(sample, priors: PriorDict):
    def cdf_op(sample, prior):
        return prior.cdf(sample)

    return _process_samples_for_priors(sample, priors, cdf_op)


def _process_samples_for_priors(sample, priors: PriorDict, op_func):
    if isinstance(sample, np.ndarray):
        shape = sample.shape
        if len(shape) == 1:
            if len(sample) == len(priors):
                return np.array([op_func(sample[i], p) for i, p in enumerate(priors.values())])
            else:
                raise ValueError("Input sample must be same size as priors")
        elif len(shape) == 2:
            if shape[1] == len(priors):
                out_arr = np.empty_like(sample)
                priors_list = list(priors.values())
                for i, psamp_set in enumerate(sample.T):
                    out_arr[:, i] = op_func(psamp_set, priors_list[i])
                return out_arr
            else:
                raise ValueError("Shape mismatch: Could not broadcast input sample to priors")
        else:
            raise ValueError(f"Invalid shape {shape} for sample")
    elif isinstance(sample, dict):
        return {k: op_func(v, priors[k]) for k, v in sample.items()}
    elif isinstance(sample, pd.Series):
        return pd.Series({k: op_func(v, priors[k]) for k, v in sample.items()})
    elif isinstance(sample, pd.DataFrame):
        out_df = pd.DataFrame(index=sample.index)
        for c in sample.columns:
            out_df[c] = op_func(sample[c].to_numpy(), priors[c])
        return out_df
    elif isinstance(sample, list):
        assert all([isinstance(subsample, dict) for subsample in sample])
        return [_process_samples_for_priors(subsample, priors, op_func) for subsample in sample]


def convert_sample_type(sample, priors, target_type: str):
    # JUST A STUB RIGHT NOW
    if target_type == SampleTypes.INPUT:
        return sample

    if isinstance(sample, np.ndarray):
        if target_type == SampleTypes.ARRAY:
            return sample
        elif target_type == SampleTypes.LIST_OF_DICTS:
            if len(sample.shape) == 1:
                sample = sample.reshape((len(sample), 1))
            if len(sample.shape) == 2:
                return [{k: subsample[i] for i, k in enumerate(priors)} for subsample in sample]
            else:
                raise ValueError("Shape mismatch: Could not broadcast input sample to priors")
        elif target_type == SampleTypes.DICT:
            assert len(sample.shape) == 1
            assert len(sample) == len(priors)
            return {k: sample[i] for i, k in enumerate(priors)}
        else:
            raise ValueError(f"Target type {target_type} not supported for array inputs")
    elif isinstance(sample, list):
        assert isinstance(sample[0], dict)
        if target_type == SampleTypes.ARRAY:
            return _lod_to_arr(sample, priors)
        elif target_type == SampleTypes.PANDAS:
            return pd.DataFrame(_lod_to_arr(sample, priors), columns=priors)
    elif isinstance(sample, pd.DataFrame):
        if target_type == SampleTypes.LIST_OF_DICTS:
            return [v.to_dict() for _, v in sample.iterrows()]
        elif target_type == SampleTypes.ARRAY:
            return sample.to_numpy()
        elif target_type == SampleTypes.PANDAS:
            return sample

    raise TypeError(
        "Unsupported combination of input type and target type", type(sample), target_type
    )


def _lod_to_arr(in_lod, priors):
    assert len(in_lod[0]) == len(priors)
    out_arr = np.empty((len(in_lod), len(in_lod[0])))
    for i, in_dict in enumerate(in_lod):
        for j, k in enumerate(priors):
            out_arr[i, j] = in_dict[k]
    return out_arr


class SampledPriorsManager:
    def __init__(self, priors):
        self.priors = priors

    def constrain(self, sample, bounds=0.99, ret_type=SampleTypes.INPUT):
        return convert_sample_type(constrain(sample, self.priors, bounds), self.priors, ret_type)

    def ppf(self, sample, ret_type=SampleTypes.INPUT):
        return convert_sample_type(ppf(sample, self.priors), self.priors, ret_type)

    def cdf(self, sample, ret_type=SampleTypes.INPUT):
        return convert_sample_type(cdf(sample, self.priors), self.priors, ret_type)

    def convert(self, sample, ret_type):
        return convert_sample_type(sample, self.priors, ret_type)



from typing import List, Dict
from dataclasses import dataclass

from summer2 import CompartmentalModel

from jax import jit
import numpy as np

import pandas as pd

from .targets import BaseTarget
from .priors import BasePrior
from .utils.sample import SampleTypes, SampledPriorsManager


@dataclass
class ResultsData:
    derived_outputs: pd.DataFrame
    extras: dict


class BayesianCompartmentalModel:
    def __init__(
        self,
        model: CompartmentalModel,
        parameters: dict,
        priors: list,
        targets: list,
        extra_ll=None,
    ):
        self.model = model

        self._model_parameters = model.get_input_parameters()

        self.parameters = parameters
        self.targets: Dict[str, BaseTarget] = _named_list_to_dict(targets)

        for t in targets:
            priors = priors + t.get_priors()
        self.priors: Dict[str, BasePrior] = _named_list_to_dict(priors)

        self._ref_idx = self.model._get_ref_idx()
        if not isinstance(self._ref_idx, pd.Index):
            self._ref_idx = pd.Index(self._ref_idx)
        self.epoch = self.model.get_epoch()

        self._build_logll_funcs(extra_ll)

        self.sample = SampledPriorsManager(self.priors)

    def _construct_targets(self, targets: list) -> Dict[str, BaseTarget]:
        tdict = {}

        for t in targets:
            if t.name in tdict:
                raise ValueError("Duplicate target name", t.name)
            else:
                tdict[t.name] = t

        return tdict

    def _build_logll_funcs(self, extra_ll=None):
        model_params = self.model.get_input_parameters()
        dyn_params = list(model_params.intersection(set(self.priors)))
        self.model.set_derived_outputs_whitelist(
            list(set([t.model_key for t in self.targets.values()]))
        )

        self._ll_runner = self.model.get_runner(
            self.parameters, dyn_params, include_full_outputs=False
        )

        self.model.set_derived_outputs_whitelist([])
        self._full_runner = self.model.get_runner(
            self.parameters, dyn_params, include_full_outputs=False
        )

        self._evaluators = {}
        for k, t in self.targets.items():
            tev = t.get_evaluator(self._ref_idx, self.epoch)
            self._evaluators[k] = tev.evaluate

        @jit
        def logll(**kwargs):
            dict_args = capture_model_kwargs(self.model, **kwargs)
            res = self._ll_runner._run_func(dict_args)["derived_outputs"]

            logdens = 0.0
            for tname, target in self.targets.items():
                model_key = target.model_key
                evaluator = self._evaluators[tname]
                modelled = res[model_key]
                logdens += evaluator(modelled, kwargs)

            if extra_ll:
                logdens += extra_ll(kwargs)

            return logdens

        logll.__doc__ = f"""logll({', '.join([k for k in self.priors])})\n
        Run the model for a given set of parameters, and 
        return the loglikelihood of its outputs, including any values from extrall"""

        @jit
        def logll_multi(modelled_do, **kwargs):
            out_ll = {}

            for tname, target in self.targets.items():
                model_key = target.model_key
                evaluator = self._evaluators[tname]
                modelled = modelled_do[model_key]
                out_ll[tname] = evaluator(modelled, kwargs)

            if extra_ll:
                out_ll["extra_ll"] = extra_ll(kwargs)

            return out_ll

        self._logll_multi = logll_multi
        self.loglikelihood = logll

    def logprior(self, **parameters):
        lp = 0.0
        for k, p in self.priors.items():
            lp += np.sum(p.logpdf(parameters[k]))
        return lp

    def logposterior(self, **parameters):
        return self.loglikelihood(**parameters) + self.logprior(**parameters)

    def run(self, parameters: dict, include_extras=False) -> ResultsData:
        """Run the model for a given set of parameters.
        Note that only parameters specified as priors affect the outputs; other parameters
        are in-filled from the initial arguments supplied to BayesianCompartmentalModel

        Args:
            parameters: Dict of parameter key/values (as specified in priors)

        Returns:
            ResultsData, an extensible container with derived_outputs as a DataFrame
        """
        run_params = {k: v for k, v in parameters.items() if k in self._model_parameters}
        results = self._full_runner._run_func(run_params)

        if include_extras:
            extras = {}
            ll_components = self._logll_multi(results["derived_outputs"], **parameters)
            extras["ll_components"] = ll_components
            extras["loglikelihood"] = sum(ll_components.values())
            extras["logprior"] = self.logprior(**parameters)
            extras["logposterior"] = extras["logprior"] + extras["loglikelihood"]
        else:
            extras = {}

        return ResultsData(
            derived_outputs=pd.DataFrame(results["derived_outputs"], index=self._ref_idx),
            extras=extras,
        )

    def run_jax(self, parameters: dict) -> dict:
        """Run the jax run function for the model directly with the supplied parameters;
        meaning bcm.run_jax can be included in JIT calls

        Args:
            parameters: Dict of parameter key/values (as specified in priors)

        Returns:
            Results as per the summer2 jax runner
        """
        return self._full_runner._run_func(parameters)


def capture_model_kwargs(model: CompartmentalModel, **kwargs) -> dict:
    model_params = model.get_input_parameters()
    return {k: kwargs[k] for k in kwargs if k in model_params}


def _named_list_to_dict(in_list: list) -> dict:
    tdict = {}

    for t in in_list:
        if t.name in tdict:
            raise ValueError("Duplicate name in input list", t.name)
        else:
            tdict[t.name] = t

    return tdict
